;; Generated package description from code-cells.el  -*- no-byte-compile: t -*-
(define-package "code-cells" "0.5.0.20250608.143915" "Lightweight notebooks with support for ipynb files" '((emacs "27.1") (compat "29.1")) :commit "b99013b81712edbb94fdb57cce083d32ad06fbb8" :authors '(("Augusto Stoffel" . "arstoffel@gmail.com")) :maintainer '("Augusto Stoffel" . "arstoffel@gmail.com") :keywords '("convenience" "outlines") :url "https://github.com/astoff/code-cells.el")
